/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * config.h:  Hardware configuration constants
 **********************************************************/

//PIC24EP64GP202 pin layout

// Pin Function         Assignment (*=shared/reconfigurable)
// ----------------------------------------------------------
// 1:  /MCLR
// 2:  RA0/AN0          ADC1        Lower manual aftertouch
// 3:  RA1/AN1          ADC1        Upper manual aftertouch
// 4:  RB0/AN2          ADC1        Expression pedal
// 5:  RB1/AN3          ADC1        Solo manual aftertouch
// 6:  RB2/RPI34        SPI2:SDI*   Control panel write data line (SDO)
// 7:  RB3/RP35         SPI2:SDI*   Control panel read data line  (SDi)
// 8:  Vss
// 9:  RA2              GPIO        Drawbar initialization pulse  (TB) - not needed
// 10: RA3              CN          Drawbar advance pulse         (TAC1)
// 11: RB4/RP36         UART1:RX    RS-232 RX
// 12: RA4/RP20         UART1:TX    RS-232 TX
// 13: Vdd
// 14: RB5/RP37         GPIO        Shift register shift/load signal, status LED
// 15: RB6/RP38         UART2:TX    MIDI TX
// 16: RB7/RP39/SCK1    SPI1:SCK    Shift register clock
// 17: RB8/RP40/T4CK    INT2        Control panel sync signal (C1), inverted
// 18: RB9/RP41/SDI1    SPI1:SDI    Shift register data
// 19: Vss
// 20: Vcap
// 21: RB10/RP42        SPI2:SCK    Control panel clock (SYW/M), inverted
// 22: RB11/RP43        INT1*       Low C driver sense, inverted
// 23: RB12/RPI44       INT1*       A# driver sense, inverted
// 24: RB13/RPI45       INT1*       F driver sense, inverted
// 25: RB14/RPI46/T3CK  SPI2:SDI*   Drawbar read data line (TDi)
// 26: RB15/RPI47/T5CK  CN          Control panel shift enable signal (C0)
// 27: AVss
// 28: AVdd

//Interrupt priorities:
// INT1 = 7     kbscan.c    Keyboard scan sync
// T5   = 6     kbscan.c    Keyboard scan read timer
// DMA0 = 5     shiftreg.c  Keyboard scan read complete
// INT2 = 4     cpanel.c    Control panel C1 sync signal
// CN   = 4     cpanel.c    Control panel C0 changed or drawbar TAC1 changed
// U1TX = 2     uart.c      RS-232 transmit data complete
// U1RX = 1     uart.c      RS-232 receive data complete
// T1   = 1     debug.c     Periodic debug timer

// GPIO-related constants

//AN0..3 analog, the rest digital
//Outputs on:
//  * RA4 (RS232; peripheral-controlled)
//  * RB5 (Shift register mode / LED)
//  * RB6 (MIDI; peripheral-controlled)
//  * RB7 (Shift register clock)
#define CONFIG_ANSELA 3      //RA0,1 analog
#define CONFIG_ANSELB 3      //RB0,1 analog
#define CONFIG_TRISA 0xFFEF  //Output on RA4
#define CONFIG_TRISB 0xFF1F  //Output on RB5,6,7
#define LED_ON 0x20 //LATB

// RS-232 communications

#define UART_RX_REMAP_PIN 36
#define UART_TX_REMAP_PIN_REG _RP20R
#define U1TX 1
#define UART_TX_ON 0x0010 //LATA

//MIDI

#define MIDI_TX_REMAP_PIN_REG _RP38R
#define U2TX 3

//Keyboard scanning
#define FX_KBSCAN_CL        _RB11       //Low C scan phase driver - Active high, variable up to 12kHz
#define FX_KBSCAN_CL_RPI    43
#define FX_KBSCAN_AS        _RB12       //A# scan phase driver - Active high, variable up to 12kHz
#define FX_KBSCAN_AS_RPI    44
#define FX_KBSCAN_F         _RB13       //F scan phase driver - Active low, variable up to 12kHz
#define FX_KBSCAN_F_RPI     45

//Control panel
#define FX_CPANEL_C0        _RB15       //Control bit 0
#define FX_CPANEL_C1        _RB8        //Control bit 1 (inverted)
#define FX_CPANEL_SYW       _RB10       //Clock (20kHz, inverted)
//define FX_CPANEL_SDO      _RB2        //LED/motor data out to panel
//define FX_CPANEL_SDI      _RB3        //Button/slider data in from panel
#define FX_CPANEL_TAC1      _RA3
#define FX_CPANEL_TB        _RA2
//define FX_CPANEL_TDI      _RB14

#define FX_CPANEL_C0_RPI    47
#define FX_CPANEL_C1_RPI    40
#define FX_CPANEL_SYW_RPI   42
#define FX_CPANEL_SDO_RPI   34
#define FX_CPANEL_SDI_RPI   35
#define FX_CPANEL_TDI_RPI   46

//PIC special function registers (omissions or header differences between C30 and XC16)
#define _SPI2EN SPI2STATbits.SPIEN

#ifndef _SCK2INR
#define _SCK2INR _SCK2R
#endif
