/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012-2013
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * cpanel.c:  Reading the control panel and drawbars
 **********************************************************/
#include "common.h"

/*
FX-20 control panel timing:

All signals are to be read on SYW rising edge.
Note: In practice, it doesn't seem to matter if read on rising or falling edge of SYW

                                Read buttons/sliders                                         Write LEDs/sliders                                         Read drawbars
        ____________________________________________________________________   ___________________________________________  _______________________________________________________________________
       /                                                                    \ /                                           \/                                                                       \

SYW:  ..._#_#_#_#_#_#_#_#_#_#_#_..._#_#_#_#_#_#_#_#_#_..._#_#_#_#_#_#_#_#_#_..._#_#_#_#_#_#_#_#_#_..._#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_..._#_#_#_#_#_#_#_#_#_#_..._#_#_#_#_#_#_#_#_#_#_...
C1:   ...___##__________________...___________________...___________________...___________________..._____________________##______________________..._____________________..._____________________...
C0:   ..._____################__..._################__..._################__..._################__..._################____##______________________..._____________________..._____________________...
SDO:  ..._________________##____..._____________##____..._____________##____..._h g f e d c b a __..._h g f e d c b a ____________________________..._____________________..._____________________...
SDi:  ..._____h g f e d c b a   ..._h g f e d c b a   ..._____________##____..._____________##____..._____________##______________________________..._____________________..._____________________...
TAC1: ..._____________________##...###################...###################...##################_...##################___######__________________..._##__________________..._##__________________...
TB:   ..._______________________...___________________...___________________...___________________..._____________________##______________________..._____________________..._____________________...
TDi:  ..._______________________...___________________...___________________...___________________...___________________####____h g f e d c b a __...___h g f e d c b a __..._____######__######__...

                                 \_____________________/                     \_____________________/                                               \_______________________/
                                          x56                                          x56                                                                    x13

Phase 1: Read buttons/sliders
    * Setup: C1:C0 = 10
    * Data transfer: 58 bytes, SPI-compatible protocol (8-bit) using:
       > SYW = SCK
       > C0  = SS (inverted from normal SPI use: logic 1 indicates active)
       > SDi = SData, MSb first
    * 58th data byte is 0x02

Phase 2: Write LEDs/sliders
    * Data transfer: 57 bytes, SPI-compatible protocol (8-bit) using:
       > SYW = SCK
       > C0 = SS (inverted from normal SPI use: logic 1 indicates active)
       > SDO = SData, MSb first
    * Stop bit: C1:C0 = 11

Phase 3: Read drawbars (not implemented in this firmware yet)
    * Data transfer: 15 bytes, SPI-compatible protocol (8-bit framed mode) using:
       > SYW = SCK
       > TAC1 = SS frame pulse (inverted from normal SPI use: logic 1 indicates active)
       > TDi = SData, MSb first, 8-bit frames
    * 15th data byte is 0x77


Note: SYW, C1 are inverted before PIC inputs, so module should be set for:
  Read signals on /SYW falling edge
  Enable SPI read when C0 is high (SDO,SDi) or falls (TDi)

On /C1 fall to 0 (INT2):
    Wait for /SYW == 0
    If /C1==0
        If C0==0:
            Start 57-byte DMA read from SDi
        Else
            Reset DMA to TDi, start listening to TAC1

On DMA read from SDi complete:
    Start 57-byte DMA read from SDO

On TAC1 fall to 0 (CN, TAC1==0):
    Immediately start 8-bit DMA read from TDi
    Stop listening to TAC1 and start listening to C0 after 14th rise.


SPI2 MODULE CONFIGURATION
=========================
Mode:   8-bit slave mode (clock is input), read data on falling edge of /SYW
        Disable /SS pin                                         [CKP=1,SSEN=0]

Control bits:
    DISSCK = 0      (N/A for slave mode)
    DISSDO = 1      SDO pin not used
    MODE16 = 0      8-bit mode
    SMP = 0         (required for slave mode)
    CKE = 0         (not using output pin)
    SSEN = 0        (cannot use slave select pin for C0 signal because of incorrect polarity)
    CKP = 1         Clock idle is high (input data sampled on transition to idle)
    MSTEN = 0       Slave mode
    SPRE = 000      (N/A for slave mode)
    PPRE = 000      (N/A for slave mode)

    FRMEN = 0       Framed mode disabled
    SPIFSD = 0      (N/A when framed mode disabled)
    FRMPOL = 0      (N/A when framed mode disabled)
    FRMDLY = 0      (N/A when framed mode disabled)
    SPIBEN = 0      No enhanced buffer
*/

//==========================================================
// Globals
//==========================================================
uint16 fx_cpanel_bitmask[65];               //Bitmask (AND mask) indicating which bits to read: [0]..[28] = buttons, [29]..[57]=LEDs, [58]..[64]=drawbars
volatile uint16 cpanel_buffer[65];          //Masked data read from control panel and drawbars: [0]..[28] = buttons, [29]..[57]=LEDs, [58]..[64]=drawbars
volatile bool fx_cpanel_changed = false;    //Flag set when the eavesdropping system detects a change in the masked bits

//==========================================================
// Locals
//==========================================================
volatile static byte* read_ptr = (byte*)cpanel_buffer;      //Pointer to current buffer byte
volatile static byte* mask_ptr = (byte*)fx_cpanel_bitmask;  //Pointer to current bitmask byte
volatile static uint16 read_count = 0;                      //How many bytes left to read

//==========================================================
// cpanel_init()
//  Configures and starts the control panel eavesdropping
//  system.  Uses peripherals: SPI2, INT2, CN
//==========================================================
void cpanel_init() {
    //Disable SPI2 for now
    _SPI2EN = 0;
    _SCK2INR = FX_CPANEL_SYW_RPI;
    SPI2CON1 = 0x0840;  //8-bit slave mode, /SS pin disabled

    //Configure and disable CN interrupt
    _CNIEA3 = 0;
    _CNIEB15 = 0;
    _CNIP = 4;
    _CNIF = 0;

    //Configure and enable INT2
    _INT2EP = 1;    //Interrupt on falling edge of /C1
    _INT2IP = 4;
    _INT2IF = 0;
    _INT2R = FX_CPANEL_C1_RPI;
    _INT2IE = 1;

    //Enable CN interrupt
    _CNIE = 1;
}

//==========================================================
// reconfigure_for_sdi()
//  Reinitializes system to start of phase 1 (reading SDi)
//==========================================================
static void reconfigure_for_sdi() {
    _SPI2EN = 0;
    read_ptr = (byte*)cpanel_buffer;
    mask_ptr = (byte*)fx_cpanel_bitmask;
    read_count = 58;
    _SDI2R = FX_CPANEL_SDI_RPI;
    _CNIEB15 = 1;
}

//==========================================================
// Interrupt handler: INT2
// Triggered on /C1 = 0 (i.e. C1==1)
// Resynchronizes control panel read to SDi or TDi
//==========================================================
void __attribute__((interrupt,no_auto_psv)) _INT2Interrupt() {
    _INT2IF = 0;

    //Wait for clock falling edge (13us typ.)
    while(FX_CPANEL_SYW == 1) { }

    if(FX_CPANEL_C1==0) {
        if(FX_CPANEL_C0==0) {
            //C1 high, C0 low => YM2121 Load, just before transferring SDi followed by SDO
            //Begin 57-byte read from SDI
            reconfigure_for_sdi();
        } else {
            //C1 high, C0 high => YM2121 Latch, just before transferring TDi
            //@@@ Begin drawbar read sequence
        }
    }
}

//==========================================================
// Interrupt handler: CN
// Triggered on C0 change
// On C0 rising edge:
//      Begin 8-bit SPI2 read
// On C0 falling edge:
//      Wait for SPI2 read to finish, then transfer data
//==========================================================
void __attribute__((interrupt,no_auto_psv)) _CNInterrupt() {
    _CNIF = 0;
    if(FX_CPANEL_C0 == 1) {
        //C0 enabled - begin shifting one byte
        SPI2STATbits.SPIROV = 0;
        _SPI2EN = 1;
        SPI2BUF = 0;
    } else {
        //C0 disabled - Wait until hardware finishes, then store the byte
        while(!SPI2STATbits.SPIRBF) { }
        if(read_count > 0) {
            //Mask out any bits we're not listening to, and flag changes
            byte data = (SPI2BUF & (*mask_ptr)) ^ (*read_ptr);
            if(data != 0) {
                (*read_ptr) ^= data;
                fx_cpanel_changed = true;
            }

            //Advance to the next byte
            ++read_ptr;
            ++mask_ptr;
            --read_count;

            //Advance to the next input line
            if(!read_count && _SDI2R == FX_CPANEL_SDI_RPI) {
                //Finished reading SDi, now read SDO
                read_count = 57;
                _SDI2R = FX_CPANEL_SDO_RPI;
            }
        }
        _SPI2EN = 0;
    }
}
