/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * kbscan.c:  Keyboard scanning
 *
 * Basic MIDI (no velocity sensing):
 * Tier 1 interrupt (CL/A#/F#):
 *  - Synchronizes read timer to scan phase
 * Tier 2 interrupt (read timer):
 *  - Loads shift register from line
 *  - Saves pitch data
 *  - Begins 16-bit shift register read
 * Tier 3 interrupt (shift register read finished):
 *  - Saves value from shift register in key state buffer
 **********************************************************/
#include "common.h"
#include "shiftreg.h"
#include "debug.h"
#include "uart.h"

//Configuration for free-running timer TMR4:
//      Note: 156kHz * 385 ~= 60MHz;  12kHz * 5000 = 60MHz
//      T4CON = 1000 0000 0000 0000 (On, prescale 1:1, 16-bit mode, from internal clock Fp)

//Configuration for CL interrupt (INT1):
//      Pin select register: RPINR0
//      Interrupt control:   _INT1EP (INTCON2<1>) = Edge polarity; 0=rising, 1=falling
//                           _INT1IF, _INT1IE, _INT1IP
//      Interrupt routine:   _INT0Interrupt()

//Configuration for phase timer TMR5:
//      T5CON = 1000 0000 0000 0000 (On, prescale 1:1, from internal clock Fp)
//      PR5 = (Measured TMR4 Interval)/13
//      TMR5 = PR5/2
//      T5CON = 0000 0000 0000 0000 (On, prescale 1:1, from internal clock Fp)
//      Interrupt control:  _T5IF, _T5IE, _T5IP
//      Interrupt routine:  _SPI1Interrupt()

//Configuration for shift register non-blocking read:
//      Interrupt control:  _SPI1IF, _SPI1IE, _SPI1IP
//      Interrupt routine:  _SPI1Interrupt()

//Algorithm:
// INT0:  Sync timer
// T5:    Store RB value, Start shift register DMA transfer
// DMA0:
//  - Read pitch (stored RB value, four bits of shifted data; look up from array)
//  - Read active return lines (combine two words)
//  - Write active return lines to real-time buffer

//==========================================================
// Externs
//==========================================================
extern volatile uint16 key_state[13];   //defined in midi_scan.c

//==========================================================
// Locals
//==========================================================
static volatile uint16 saved_portb;     //The value of PORTB read at the same time as the shift register
static uint16 scan_phase;               //Which phase of the scan (0..2)

//==========================================================
// Local constants
//==========================================================
static const uint16 scan_phase_pin[] = { FX_KBSCAN_CL_RPI, FX_KBSCAN_F_RPI, FX_KBSCAN_AS_RPI };
static const byte pitch_map[] = {
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,     //Invalid inputs from priority encoder
    11, 6, 4, 3, 2, 1, 7, 8, 9,             //Valid inputs from priority encoder
    12, 0, 10, 0xFF, 5, 0xFF, 0xFF, 0xFF    //Valid and invalid inputs from saved portb
};

//==========================================================
// kbscan_init()
// Configures and starts the keyboard scanner.
// Peripherals used: SPI1/DMA0, INT1, T4, T5
//==========================================================
void kbscan_init() {
    saved_portb = 0;

    //Disable the shift register interrupt
    _SPI1IE = 0;

    //Initialize the timer 5 interrupt
    T5CON = 0;
    _T5IP = 6;
    _T5IF = 0;
    _T5IE = 1;

    //Initialize the CL interrupt
    _INT1EP = 0;
    _INT1IP = 7;
    _INT1IF = 0;
    scan_phase = 0;
    _INT1R = FX_KBSCAN_CL_RPI;
    _INT1IE = 1;

    //Start the free-running CL timer
    TMR4 = 0;
    T4CON = 0x8000;
}

//==========================================================
// Interrupt handler: INT1
// Triggered on direct pitch pin falling edge
// Resynchronizes keyboard scan timer
//==========================================================
void __attribute__((interrupt,no_auto_psv)) _INT1Interrupt() {
    //Advance the scan phase
    ++scan_phase;
    if(scan_phase == 3) scan_phase = 0;
    _INT1R = scan_phase_pin[scan_phase];

    //Reset state and sync the read timer
    _T5IF = 0;
    _T5IE = 1;

    PR5 = 385;
    TMR5 = 128;
    T5CON = 0x8000;

    //Clear the interrupt flag
    _INT1IF = 0;
}

//==========================================================
// Interrupt handler: T5
// Triggered at ~156kHz
// Samples keyboard scan lines and begins shifting the data
//==========================================================
void __attribute__((interrupt,no_auto_psv)) _T5Interrupt() {
    _T5IF = 0;
    saved_portb = shiftreg_load_beginread_dma();
}

//==========================================================
// Interrupt handler: DMA0
// Triggered when data read completes
// Processes and stores the key data into the buffer
//==========================================================
void __attribute__((interrupt,no_auto_psv,shadow)) _DMA0Interrupt() {
    _DMA0IF = 0;

    //Figure pitch
    byte pitch = ((spi_dma_buffer[0] & 1) << 3) | (spi_dma_buffer[1] & 7);

    if(pitch == 0xF) {
        //Pitch from PORTB pins (CL, C, F or A#)
        pitch = pitch_map[((saved_portb >> 11) & 7) + 15];
    } else if((saved_portb&0x3800) == 0) {
        //Pitch from priority encoder
        pitch = pitch_map[pitch];
    } else {
        //Invalid bit pattern
        return;
    }

    if(pitch < 13) {
        //Valid pitch; figure return lines

        //First word is key-down mask (normally set)
        //Second word is key-up mask (normally clear), except for pedals
        //Key is down if key-down bit is set or key-up bit is clear; fudge pedals
        uint16 s = key_state[pitch] | 6;    //Trust me - setting the pedal bits here works
        uint16 d = spi_dma_buffer[0] & 0xfffe;
        uint16 u = spi_dma_buffer[1] & 0xfff8;

        key_state[pitch] = s ^ ((s ^ u) & (d ^ u)); //Flip S if S!=U && D!=U
            //Old value with no upper contact debounce: ~(spi_dma_buffer[0] | 1) | (spi_dma_buffer[1] & 0xFFF8);
    }
}
