/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * logic_probe.c:  8-channel logic probe
 * 
 * NOTE: Configuration is for first revision of board and
 *       may damage the second revision.
 *
 * Saves as 3000-byte RLE data
 * First byte is initial state.
 * Additional bytes are time delta (integer ticks at 234375 Hz) and changed line number
 * 0ttttppp:  0..15 ticks delay (down to 15.6kHz), pin ppp changed
 * 1tttttttt 0ttttppp:  16..2047 ticks delay (down to 114.5 Hz), pin ppp changed
 * 1tttttttt 1ttttttt 0ttttppp: 2048..262143 ticks delay (down to 0.89 Hz), pin ppp changed
 * 1tttttttt 1ttttttt 1ttttttt 0ttttppp: 262143..33554432 ticks delay, pin ppp changed
 * 
 * Pins to watch:
 *         111111
 *         5432109876543210
 *  PORTA: ...........###..
 *  PORTB: ...#.#.....###..
 **********************************************************/
#include "common.h"
#include "uart.h"

volatile bool logic_probe_cancel;

uint16 get_pin_state() {
    uint16 a = (PORTA & 0x1C) >> 2;		// ...........###.. => .............###
    uint16 b = (PORTB & 0x141C) << 1;	// ...#.#.....###.. => ..#.#.....###...
    uint16 c = (b & 0x2000) >> 6;		// ..#............. => ........#.......
    uint16 d = (b & 0x800) >> 5;		// ....#........... => .........#......
    return (a|b|c|d)&0xFF;
}

static uint16 pos;
static byte data[3000];

static bool save_high_bits(byte b, bool force) {
    b &= 0x7F;
	if(b || force) {
        data[pos++] = b | 0x80;
		return true;
    }
	return false;
}

void logic_probe() {
    uint16 i;
    uint16 last_pin_state = get_pin_state();
    uint16* wdata = (uint16*)data;

    //Write initial pin state
    data[0] = last_pin_state;
    pos = 1;

    //Reset and start timer
    PR2 = PR3 = 0xFFFF;
    T2CON = 0x8008; //32-bit mode, 1:64 prescale, on
    TMR3HLD = 0;
    TMR2 = 0;
    
    logic_probe_cancel = false;
    while(pos <= 2996 && !logic_probe_cancel) {
        //Read the pin state
        uint16 new_pin_state = get_pin_state();
        uint16 delta_pin_state = last_pin_state ^ new_pin_state;

        //If anything has changed, save a transition
        if(delta_pin_state) {
            //Read and reset the timer
            uint16 timer_interval_low = TMR2;
            uint16 timer_interval_high = TMR3HLD;
            TMR3HLD = 0;
            TMR2 = 0;

            //Find the first set bit and save it with the interval
            for(i = 0; i < 8; ++i) {
                if(delta_pin_state & (1<<i)) {
                    bool b = save_high_bits(timer_interval_high >> 2, false);
                    b = save_high_bits(((timer_interval_low >> 11)&0x1F) | (timer_interval_high << 5), b);
                    save_high_bits(timer_interval_low >> 4, b);
                    data[pos++] = (i | (timer_interval_low << 3)) & 0x7F;
                    break;
                }
            }

            //Save all other set bits with zero delay
            for(++i; i < 8; ++i) {
                if(delta_pin_state & (1<<i)) {
                    data[pos++] = i;
                }
            }
        }
		last_pin_state = new_pin_state;
    }

    for(i = 0; i < pos; i += 2) {
        uart_print_hex(*wdata);
        uart_send_string(" ");
        ++wdata;
    }
}
