/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 *
 * main.c:  Initialization
 *
 * Supported Electone models (tested):
 *  FX-20
 *
 * Theoretically supported Electone models (incomplete/untested):
 *  FS-50, FS-70, FS-300, FS-500, FX-10
 *  FS-20, FS-30, FS-100, FS-200
 *
 * Unsupported Electone models:
 *  FX-1 (hardware may short this model)
 *  FX-3 (no available tech info)
 **********************************************************/
#include "common.h"
#include "blink_led.h"
#include "midi_out.h"
#include "uart.h"
#include "shiftreg.h"
#include "analog.h"
#include "kbscan.h"
#include "electone.h"
#include "midi_scan.h"
#include "debug.h"
#include "cpanel.h"

// Configuration registers

_FICD(ICS_PGD2 & JTAGEN_OFF)
_FPOR(ALTI2C1_OFF & ALTI2C2_OFF & WDTWIN_WIN25)
_FWDT(WDTPOST_PS32768 & WDTPRE_PR128 & PLLKEN_OFF & WINDIS_OFF & FWDTEN_OFF)
_FOSC(POSCMD_NONE & OSCIOFNC_ON & IOL1WAY_OFF & FCKSM_CSECMD)
_FOSCSEL(FNOSC_FRC & IESO_OFF)
_FGS(GWRP_OFF & GCP_OFF)

void init() {

	//Initialize I/O pins
	ANSELA = CONFIG_ANSELA;
	ANSELB = CONFIG_ANSELB;
	LATA = UART_TX_ON;
	LATB = 0x0000;
	TRISA = CONFIG_TRISA;
	TRISB = CONFIG_TRISB;

	//Set shift register to shift mode, turn LED on.
	LATB = LED_ON;

	//Delay (probably not necessary)
	unsigned int i;
	unsigned int j;
	for(i = 0; i < 60000; i++) { Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); }
	for(i = 0; i < 60000; i++) { Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); Nop(); }

	//Init oscillator to 60MHz

	//Configure PLL prescaler, PLL postscaler, PLL divisor
	PLLFBD=63; // M=65
	CLKDIVbits.PLLPOST=0; // N2=2 (120MHz * 2 = 240MHz)
	CLKDIVbits.PLLPRE=0; // N1=2 (7.37/2 = 3.69MHz)

	//Check:
	// Fosc = (M*Fin) / (N1 * N2)
	// Fosc = (33 * 7.37) / 4 = 60.8025

	//Initiate Clock Switch to FRC oscillator with PLL (NOSC=0b001)
	__builtin_write_OSCCONH(0x01);
	__builtin_write_OSCCONL(OSCCON | 0x01);

	//Wait for clock switch to occur
	while (OSCCONbits.COSC!= 0b001);

	//Wait for PLL to lock
	while (OSCCONbits.LOCK!= 1);

	//Delay (probably not necessary)
	for(i = 0; i < 50; i++) {
		for(j = 0; j < 60000; j++) { Nop(); Nop(); Nop(); Nop(); }
	}
}

extern volatile int command;

int main(void) {
	init();
	midi_init();
    uart_init();
	shiftreg_init();
	analog_init();
	debug_init();

	midi_scan_init();   //Call before initializing keyboard scan and control panel read
    kbscan_init();
	cpanel_init();

	uart_send_string("\r\nInitialized\r\n");

	/*
    command = 0;
    while(true) {
        while(!command) {}
        
        //switch(command) {
        //    case 'p':
        //        logic_probe();
        //}

        command = 0;
    }*/

	midi_scan();

	return 0;
}
