/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012-2013
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * midi_config.c:  Dynamic MIDI configuration
 **********************************************************/
#include "common.h"
#include "electone.h"
#include "midi_config.h"
#include "cpanel.h"
#include "midi_scan.h"

//==========================================================
// Default configuration for FX-20, designed for the
//  Burea Church Extended sample set for GrandOrgue
//==========================================================
const electone_config fx20_default_config = {
    //Order:
    // Pedal 210 (Note: Pedal octave 2 is solely for III:P16 coupler)
    // Lower 04321
    // Upper 12340
    // Solo 432, transposed down 1 octave

    { 1,1,1,    2,2,2,2,2,      0,0,0,0,0,      3,3,3    },  //MIDI channels per octave
    { 60,48,36, 36,84,72,60,48, 48,60,72,84,36, 72,60,48 },  //Middle C per octave
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 },            //MIDI pitch scan order
    {
        { 0xD3, 0, 1000 },                                   //Analag controls
        { 0xD2, 0, 1000 },
        { 0xD0, 0, 1000 },
        { 0xB0, 7, 800 }
    },
    {                                                        //Control panel map
        { fx20_control_footswitch, type_button | subtype_persistent, 0xC4, 38, 0 },   //Footswitch: Preset #9

        //Lower special (positive reed stops)
        { fx20_control_lspec_electricpiano, type_button | subtype_persistent, 0xC4, 88, 0 }, //Electric Piano: Krumhorn 8'
        { fx20_control_lspec_piano,         type_button | subtype_persistent, 0xC4, 87, 0 }, //Piano:          Fagott 16'
        //Lower orchestra (positive mixture, wood and odd stops)
        { fx20_control_lorch_reed,          type_button | subtype_persistent, 0xC4, 86, 0 }, //Reed:           Cymbel II
        { fx20_control_lorch_spice2,        type_button | subtype_persistent, 0xC4, 84, 0 }, //Spice 2:        Septima 1 1/7'
        { fx20_control_lorch_brass1,        type_button | subtype_persistent, 0xC4, 81, 0 }, //Brass 1:        Flojtlein 2'
        { fx20_control_lorch_spice1,        type_button | subtype_persistent, 0xC4, 83, 0 }, //Spice 1:        Nasat 1 1/3'
        { fx20_control_lorch_strings3,      type_button | subtype_persistent, 0xC4, 78, 0 }, //Strings 3:      Koppelflojt 4'
        { fx20_control_lorch_vocal2,        type_button | subtype_persistent, 0xC4, 82, 0 }, //Vocal 2:        Ters 1 3/5'
        { fx20_control_lorch_strings2,      type_button | subtype_persistent, 0xC4, 76, 0 }, //Strings 2:      Gedakt 8'
        { fx20_control_lorch_vocal1,        type_button | subtype_persistent, 0xC4, 79, 0 }, //Vocal 1:        Rorkvint 2 2/3'
        { fx20_control_lorch_strings1,      type_button | subtype_persistent, 0xC4, 73, 0 }, //Strings 1:      Gedakt 16'
        // Lower combination (positive principal stops)
        { fx20_control_lcombi_memory3,      type_button | subtype_persistent, 0xC4, 85, 0 }, //Memory 3:       Oktava 1'
        { fx20_control_lcombi_memory2,      type_button | subtype_persistent, 0xC4, 80, 0 }, //Memory 2:       Principal 2'
        { fx20_control_lcombi_memory1,      type_button | subtype_persistent, 0xC4, 77, 0 }, //Memory 1:       Gemshorn 4'
        { fx20_control_lcombi_preset1,      type_button | subtype_persistent, 0xC4, 75, 0 }, //Preset 1:       Halflojt 8'
        { fx20_control_lcombi_combilever,   type_button | subtype_persistent, 0xC4, 74, 0 }, //Combi Lever:    Gegen Principal 8'

        // Play Assist (couplers)
        { fx20_control_playassist_duet,             type_button | subtype_persistent, 0xC4, 32, 0 }, //Duet:                II/P (pedal to positive)
        { fx20_control_playassist_melodyonchord2,   type_button | subtype_persistent, 0xC4, 31, 0 }, //Melody on Chord 2:   I/P (pedal to great)
        { fx20_control_playassist_melodyonchord1,   type_button | subtype_persistent, 0xC4, 33, 0 }, //Melody on Chord 1:   III/P (pedal to swell)
        { fx20_control_playassist_uspec,            type_button | subtype_persistent, 0xC4, 34, 0 }, //Upper Special:       II/I (great to positive)
        { fx20_control_playassist_uorch,            type_button | subtype_persistent, 0xC4, 72, 0 }, //Upper Orchestra:     I 4' (great to great 8va)
        { fx20_control_playassist_ucombi,           type_button | subtype_persistent, 0xC4, 35, 0 }, //Upper Combination:   III/I (great to swell)
        { fx20_control_playassist_keyset,           type_button | subtype_persistent, 0xC4, 36, 0 }, //Key set:             III/II (positive to swell)

        //Upper/lower custom voices (pedal stops)
        { fx20_control_ulcust_touchtone,      type_button | subtype_persistent, 0xC4, 49, 0 }, //Touch Tone:      Rauschpfeife IV
        { fx20_control_ulcust_hawaiianguitar, type_button | subtype_persistent, 0xC4, 55, 0 }, //Hawaiian Guitar: Trumpet 4'
        { fx20_control_ulcust_touchvibrato,   type_button | subtype_persistent, 0xC4, 54, 0 }, //Touch Vibrato:   Krumhorn 8'
        { fx20_control_ulcust_jazzguitar,     type_button | subtype_persistent, 0xC4, 53, 0 }, //Jazz Guitar:     Trumpet 8'
        { fx20_control_ulcust_presetvibrato,  type_button | subtype_persistent, 0xC4, 52, 0 }, //Preset Vibrato:  Dulcian 16'
        { fx20_control_ulcust_violin,         type_button | subtype_persistent, 0xC4, 51, 0 }, //Violin:          Basun 16'
        { fx20_control_ulcust_cosmic4,        type_button | subtype_persistent, 0xC4, 50, 0 }, //Cosmic 4:        Kontrabasun 32'
        { fx20_control_ulcust_trombone,       type_button | subtype_persistent, 0xC4, 48, 0 }, //Trombone:        Nachthorn 2'
        { fx20_control_ulcust_cosmic3,        type_button | subtype_persistent, 0xC4, 47, 0 }, //Cosmic 3:        Flojt 4'
        { fx20_control_ulcust_trumpet,        type_button | subtype_persistent, 0xC4, 46, 0 }, //Trumpet:         Oktava 4'
        { fx20_control_ulcust_cosmic2,        type_button | subtype_persistent, 0xC4, 47, 0 }, //Cosmic 2:        Flojt 4'
        { fx20_control_ulcust_saxophone,      type_button | subtype_persistent, 0xC4, 44, 0 }, //Saxophone:       Violoncell 8'
        { fx20_control_ulcust_cosmic1,        type_button | subtype_persistent, 0xC4, 45, 0 }, //Cosmic 1:        Gedakt 8'
        { fx20_control_ulcust_clarinet,       type_button | subtype_persistent, 0xC4, 43, 0 }, //Clarinet:        Principal 8'
        { fx20_control_ulcust_vocal,          type_button | subtype_persistent, 0xC4, 41, 0 }, //Vocal:           Subbas 16'
        { fx20_control_ulcust_oboe,           type_button | subtype_persistent, 0xC4, 42, 0 }, //Oboe:            Violon 16'
        { fx20_control_ulcust_bassguitar,     type_button | subtype_persistent, 0xC4, 39, 0 }, //Bass Guitar:     Untersatz 32'
        { fx20_control_ulcust_flute,          type_button | subtype_persistent, 0xC4, 40, 0 }, //Flute:           Principal 16'

        //Upper special (great reed stops)
        { fx20_control_uspec_celeste,     type_button | subtype_persistent, 0xC4, 70, 0 }, //Celeste:     Trumpet 4'
        { fx20_control_uspec_harpsichord, type_button | subtype_persistent, 0xC4, 69, 0 }, //Harpsichord: Trumpet 8'
        { fx20_control_uspec_piano,       type_button | subtype_persistent, 0xC4, 68, 0 }, //Harp:        Trumpet 16'
        //Upper orchestra (great mixture and wood stops)
        { fx20_control_uorch_brass3,      type_button | subtype_persistent, 0xC4, 67, 0 }, //Brass 3:     Mixtur V
        { fx20_control_uorch_brass2,      type_button | subtype_persistent, 0xC4, 66, 0 }, //Brass 2:     Sesquialtera II
        { fx20_control_uorch_strings4,    type_button | subtype_persistent, 0xC4, 65, 0 }, //Strings 4:   Flojt 2'
        { fx20_control_uorch_strings3,    type_button | subtype_persistent, 0xC4, 62, 0 }, //Strings 3:   Rorflojt 4'
        { fx20_control_uorch_strings2,    type_button | subtype_persistent, 0xC4, 58, 0 }, //Strings 2:   Quintadena 8'
        { fx20_control_uorch_strings1,    type_button | subtype_persistent, 0xC4, 60, 0 }, //Strings 1:   Gedackt 8'
        //Upper combination (great principal and string stops)
        { fx20_control_ucombi_memory3,    type_button | subtype_persistent, 0xC4, 64, 0 }, //Memory 3:    Oktava 2'
        { fx20_control_ucombi_preset3,    type_button | subtype_persistent, 0xC4, 63, 0 }, //Preset 3:    Gambette 4'
        { fx20_control_ucombi_memory2,    type_button | subtype_persistent, 0xC4, 61, 0 }, //Memory 2:    Oktava 4'
        { fx20_control_ucombi_preset2,    type_button | subtype_persistent, 0xC4, 59, 0 }, //Preset 2:    Gamba 8'
        { fx20_control_ucombi_memory1,    type_button | subtype_persistent, 0xC4, 57, 0 }, //Memory 1:    Principal 8'
        { fx20_control_ucombi_preset1,    type_button | subtype_persistent, 0xC4, 56, 0 }, //Preset 1:    Borduna 16'

        //Auto rhythm selection (general pistons 11-30)
        { fx20_control_autorhythm_4,            type_button | subtype_persistent, 0xC4, 20, 0 }, //Pattern 4:   GP20
        { fx20_control_autorhythm_2,            type_button | subtype_persistent, 0xC4, 30, 0 }, //Pattern 3:   GP30
        { fx20_control_autorhythm_3,            type_button | subtype_persistent, 0xC4, 19, 0 }, //Pattern 2:   GP19
        { fx20_control_autorhythm_1,            type_button | subtype_persistent, 0xC4, 29, 0 }, //Pattern 1:   GP29
        { fx20_control_autorhythm_16beat,       type_button | subtype_persistent, 0xC4, 18, 0 }, //16-beat:     GP18
        { fx20_control_autorhythm_8beat2,       type_button | subtype_persistent, 0xC4, 28, 0 }, //8-beat 2:    GP28
        { fx20_control_autorhythm_disco,        type_button | subtype_persistent, 0xC4, 17, 0 }, //Disco:       GP17
        { fx20_control_autorhythm_8beat1,       type_button | subtype_persistent, 0xC4, 27, 0 }, //8-beat 1:    GP27
        { fx20_control_autorhythm_latinrock,    type_button | subtype_persistent, 0xC4, 16, 0 }, //Latin Rock:  GP16
        { fx20_control_autorhythm_slowrock,     type_button | subtype_persistent, 0xC4, 26, 0 }, //Slow Rock:   GP26
        { fx20_control_autorhythm_samba,        type_button | subtype_persistent, 0xC4, 15, 0 }, //Samba:       GP15
        { fx20_control_autorhythm_bounce,       type_button | subtype_persistent, 0xC4, 25, 0 }, //Bounce:      GP25
        { fx20_control_autorhythm_bossanova,    type_button | subtype_persistent, 0xC4, 14, 0 }, //Bossa-Nova:  GP14
        { fx20_control_autorhythm_swing,        type_button | subtype_persistent, 0xC4, 24, 0 }, //Swing:       GP24
        { fx20_control_autorhythm_latin2,       type_button | subtype_persistent, 0xC4, 13, 0 }, //Latin 2:     GP13
        { fx20_control_autorhythm_ballad,       type_button | subtype_persistent, 0xC4, 23, 0 }, //Ballad:      GP23
        { fx20_control_autorhythm_latin1,       type_button | subtype_persistent, 0xC4, 12, 0 }, //Latin 1:     GP12
        { fx20_control_autorhythm_waltz,        type_button | subtype_persistent, 0xC4, 22, 0 }, //Waltz:       GP22
        { fx20_control_autorhythm_tango,        type_button | subtype_persistent, 0xC4, 11, 0 }, //Tango:       GP11
        { fx20_control_autorhythm_march,        type_button | subtype_persistent, 0xC4, 21, 0 }, //March:       GP21

        //Solo custom voices (swell stops)
        { fx20_control_solo_slidecontrol,  type_slider,                      0xB3, 7,   0 }, //Slide Control slider:  Swell expression 0..127
        { fx20_control_solo_touchtone,     type_button | subtype_persistent, 0xC4, 105, 0 }, //Touch Tone:     Tremulant
        { fx20_control_solo_touchvibrato,  type_button | subtype_persistent, 0xC4, 104, 0 }, //Touch Vibrato:  Skalmeja 8'
        { fx20_control_solo_presetvibrato, type_button | subtype_persistent, 0xC4, 103, 0 }, //Preset Vibrato: Dulcian 16'
        { fx20_control_solo_down,          type_button | subtype_persistent, 0xC4, 106, 0 }, //Transpose down: Coupler III 16'
        { fx20_control_solo_cosmic2,       type_button | subtype_persistent, 0xC4, 100, 0 }, //Cosmic 2:       Ters 1 3/5'
        { fx20_control_solo_horn,          type_button | subtype_persistent, 0xC4, 102, 0 }, //Horn:           Scharf III
        { fx20_control_solo_cosmic1,       type_button | subtype_persistent, 0xC4, 98,  0 }, //Cosmic 1:       Rorkvint 2 2/3'
        { fx20_control_solo_trumpet2,      type_button | subtype_persistent, 0xC4, 101, 0 }, //Trumpet 2:      Sivflojt 1'
        //{ fx20_control_solo_vocal,         type_button | subtype_persistent, 0xC4, 0, 0 }, //Vocal:          ---
        { fx20_control_solo_trumpet1,      type_button | subtype_persistent, 0xC4, 99,  0 }, //Trumpet 1:      Valdflojt 2'
        //{ fx20_control_solo_harmonica,     type_button | subtype_persistent, 0xC4, 0, 0 }, //Harmonica:      ---
        { fx20_control_solo_saxophone,     type_button | subtype_persistent, 0xC4, 96,  0 }, //Saxophone:      Halflojt 4'
        { fx20_control_solo_jazzguitar,    type_button | subtype_persistent, 0xC4, 97,  0 }, //Jazz Guitar:    Salicette 4'
        { fx20_control_solo_clarinet,      type_button | subtype_persistent, 0xC4, 95,  0 }, //Clarinet:       Principal 4'
        { fx20_control_solo_cello,         type_button | subtype_persistent, 0xC4, 94,  0 }, //Cello:          Viol Celeste 8'
        { fx20_control_solo_oboe,          type_button | subtype_persistent, 0xC4, 92,  0 }, //Oboe:           Rorflojt 8'
        { fx20_control_solo_violin,        type_button | subtype_persistent, 0xC4, 93,  0 }, //Violin:         Salicional 8'
        { fx20_control_solo_flute,         type_button | subtype_persistent, 0xC4, 91,  0 }, //Flute:          Principalflojt 8'
        { fx20_control_solo_trombone,      type_button | subtype_persistent, 0xC4, 90,  0 }, //Trombone:       Gedakt Pommer 16'
        //{ fx20_control_solo_piccolo,       type_button | subtype_persistent, 0xC4, 0, 0 }, //Piccolo:        ---

        //Preset panel (general pistons 1..10)
        { fx20_control_preset_8, type_button | subtype_persistent, 0xC4, 8,  0 }, //Preset 8
        { fx20_control_preset_7, type_button | subtype_persistent, 0xC4, 7,  0 }, //Preset 7
        { fx20_control_preset_6, type_button | subtype_persistent, 0xC4, 6,  0 }, //Preset 6
        { fx20_control_preset_5, type_button | subtype_persistent, 0xC4, 5,  0 }, //Preset 5
        { fx20_control_preset_4, type_button | subtype_persistent, 0xC4, 4,  0 }, //Preset 4
        { fx20_control_preset_3, type_button | subtype_persistent, 0xC4, 3,  0 }, //Preset 3
        { fx20_control_preset_2, type_button | subtype_persistent, 0xC4, 2,  0 }, //Preset 2
        { fx20_control_preset_1, type_button | subtype_persistent, 0xC4, 1,  0 }, //Preset 1
        { fx20_control_preset_r, type_button | subtype_persistent, 0xC4, 10, 0 }, //Preset 10
        { fx20_control_preset_d, type_button | subtype_persistent, 0xC4, 9,  0 }, //Preset 9
        { fx20_control_preset_c, type_button | subtype_persistent, 0xC4, 0,  0 }, //Clear all stops

        //Tremolo controls (tremulants)
        { fx20_control_tremolo_lcombi, type_button | subtype_persistent, 0xC4, 89, 0 }, //Lower combination: Positive tremulant
        { fx20_control_tremolo_ucombi, type_button | subtype_persistent, 0xC4, 71, 0 }, //Upper combination: Great tremulant

        //Solo keyboard LEDs
        { fx20_signal_solo_touchtone,   type_led | subtype_momentary, 0x94, 105, 0 },           //Solo Touch Tone:     Tremulant
        { fx20_signal_solo_down,        type_led | subtype_momentary, 0x94, 106, 0 },           //Solo Transpose down: Coupler III 16'
        { fx20_signal_solo_coupler,     type_led | subtype_momentary, coupler_iii_p16, 0, 0 }, //Coupler solo to upper:  III:P 16' (Pre-MIDI)

        //Tremolo LEDs
        { fx20_signal_tremolo_lcombi, type_led | subtype_momentary, 0x94, 89, 0 }, //Lower combination: Positive tremulant
        { fx20_signal_tremolo_ucombi, type_led | subtype_momentary, 0x94, 71, 0 }, //Upper combination: Great tremulant

        { 0xFFFF, 0, 0, 0, 0 }, //End record

    }
};

//==========================================================
// build_cpanel_mask_table()
//  Local function to compute the bitmask and skip vector
//  tables for the current fx_config.
//==========================================================
static void build_cpanel_mask_table() {
    uint16 i;  //Current control index or generic iterator
    uint16 j;  //Vector table index
    uint16 previous_bit_index = 0;
    uint16 previous_word_index = 0;

    for(i = 0; i < 65; ++i) {
        fx_cpanel_bitmask[i] = 0;
    }
    for(i = 0; i < 256; ++i) {
        fx_cpanel_skipvector[i] = 0;
    }
    for(i = 0; i < 256; ++i) {
        if(fx_config.control_map[i].bit_index == 0xFFFF) break;
        fx_config.control_map[i].bit_index ^= 8;
    }

    //Loop through the control definitions, setting the bits that are assigned in the config.
    j = 0;
    for(i = 0; i < 256; ++i) {
        uint16 bit_index = fx_config.control_map[i].bit_index;

        //Save the skip vectors
        uint16 word_index = bit_index >> 4;
        if(word_index > previous_word_index) {
            previous_word_index = word_index;
            while(j < i) {
                fx_cpanel_skipvector[j] = i;
                ++j;
            }
        }

        if(bit_index > (65 * 16) || (i > 0 && (bit_index ^ 8) <= previous_bit_index)) {
            break;
        }
        previous_bit_index = bit_index ^ 8;
        
        //Set the bit(s)
        byte control_type = fx_config.control_map[i].control_type & 0xF;
        uint16 b = 0x8000;                                                     //Buttons: 1 bit
        if(control_type == type_slider)        //Sliders: 4 bits
            b = 0xF000;
        else if(control_type == type_lever)    //Levers: 3 bits
            b = 0x7000;
        fx_cpanel_bitmask[word_index] |= (b >> (bit_index & 15));
    }
}

//==========================================================
// electone_load_config(...)
//  Loads a MIDI configuration into memory.
//  Must be done during startup before initializing the
//   MIDI data scanner.
//==========================================================
void electone_load_config(const electone_config* config_to_load) {
    //Copy the structure
    fx_config = *config_to_load;
    build_cpanel_mask_table();
}
