/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * midi.c:  Low-level MIDI communication
 **********************************************************/
#include "common.h"
#include "midi_out.h"

//==========================================================
// Locals
//==========================================================
static char _status;  //Active status byte (used for running status)

//==========================================================
// midi_init()
//  Configures the UART for MIDI operation
//==========================================================
void midi_init() {

    //Initialize locals
    _status = 0;

    //Initialize the hardware
    // * UART2, transmit only
    // * 8 data bits, 1 stop bit, no parity
    // * 31250 baud (60MHz / (16 * 120))
    // * No current (pin at 0V) is idle

    MIDI_TX_REMAP_PIN_REG = U2TX;       //Assign pin to UART 2
    U2BRG = 119;
    U2MODE = 0x8810;
    U2STA = 0xC400;     //Set after mode, per reference manual
}

//==========================================================
// send_byte(...)
//  Sends a byte of MIDI data
//==========================================================
static void send_byte(char data) {
    //Start the transmission
    while(U2STAbits.UTXBF) { }
    U2TXREG = data;
}

//==========================================================
// send_status_byte(...)
//  Sends the given status byte if different than the
//  previous status byte or if the queue is empty.
//==========================================================
static void send_status_byte(char data) {
    //Send if status has changed
    //Always send if queue is empty
    if(U2STAbits.TRMT || (data != _status)) {
        _status = data;
        send_byte(data);
    }
}

//==========================================================
// midi_send_message_1(...)
//  Sends a generic two-byte MIDI message
//==========================================================
void midi_send_message_1(byte message_and_channel, byte data) {
    send_status_byte(message_and_channel | 0x80);
    send_byte(data & 127);
}

//==========================================================
// midi_send_message_2(...)
//  Sends a generic three-byte MIDI message
//==========================================================
void midi_send_message_2(byte message_and_channel, byte data1, byte data2) {
    send_status_byte(message_and_channel | 0x80);
    send_byte(data1 & 127);
    send_byte(data2 & 127);
}

//==========================================================
// midi_send_note_on(...)
//  Sends a Note On message
//==========================================================
void midi_send_note_on(char channel, char pitch, char velocity) {
    send_status_byte((channel & 15) | 0x90);
    send_byte(pitch & 127);
    send_byte(velocity & 127);
}

//==========================================================
// midi_send_note_on(...)
//  Sends a Note Off message
//==========================================================
void midi_send_note_off(char channel, char pitch) {
    send_status_byte((channel & 15) | 0x90);
    send_byte(pitch & 127);
    send_byte(0);
}

//==========================================================
// midi_send_channel_aftertouch(...)
//  Sends a Channel Aftertouch message
//==========================================================
void midi_send_channel_aftertouch(char channel, char value) {
    send_status_byte((channel & 15) | 0xD0);
    send_byte(value & 127);
}

//==========================================================
// midi_send_controller(...)
//  Sends a Controller message
//==========================================================
void midi_send_controller(char channel, char controller, char value) {
    send_status_byte((channel & 15) | 0xB0);
    send_byte(controller & 127);
    send_byte(value & 127);
}
