/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * midi.h:  High-level MIDI operation
 **********************************************************/
#include "common.h"
#include "electone.h"
#include "midi_scan.h"
#include "midi_out.h"
#include "midi_config.h"
#include "cpanel.h"

//==========================================================================================
// Locals and globals for keyboard scanning
//==========================================================================================

//Scan key state buffer (stores shift register values for 13 pitches)
// Written by: Keyboard scan interrupts
volatile uint16 key_state[13];

//MIDI key state buffer (logical key state after any messages in the MIDI transmit queue)
static uint16 midi_state[13];

//Electone configuration bytes
electone_config fx_config;

//Vector table for optimization of control panel processing
byte fx_cpanel_skipvector[256];

//==========================================================================================
// Locals and globals for analog inputs
//==========================================================================================

//Which analog signal we're looking at right now.
static uint16 analog_phase = 3;

//MIDI analog hysteresis buffer
static uint16 midi_analog_hysteresis[4];
static char last_translated_value[4];

//Hysteresis window width - 20 is around 5% of 1024
#define HYSTERESIS_MAGNITUDE 40

//==========================================================================================
// Locals for control panel processing
//==========================================================================================

//MIDI control panel state buffer
static uint16 midi_cpanel_state[65];

static const byte slider_value_map[16] = { 0,0,9,18,  27,36,45,54,  64,73,82,91,  100,109,118,127 };    //Maps slider (1..15) to MIDI data range (0..127)
static const byte lever_value_map[8] = { 0,0,21,42,  64,85,106,127 };                                   //Maps lever (1..7) to MIDI data range (0..127)

//==========================================================
// midi_scan_init()
//  Initializes buffers and loads the default configuration
//==========================================================
void midi_scan_init() {
    uint16 i;
    for(i = 0; i < 13; ++i) {
        key_state[i] = 0;
        midi_state[i] = 0;
    }
    for(i = 0; i < 4; ++i) {
        midi_analog_hysteresis[i] = 0x8000;
        last_translated_value[i] = 0x80;
    }
    for(i = 0; i < 65; ++i) {
        midi_cpanel_state[i] = 0;
    }

    //Set up config for FX-20
    electone_load_config(&fx20_default_config);
}

//==========================================================
// lookup_midi_pitch(...)
//  Converts a pitch and octave index pair into a MIDI pitch
//==========================================================
static byte lookup_midi_pitch(uint16 pitch, uint16 octave) {
    return fx_config.base_pitch[octave] + fx_config.pitch_map[pitch];
}

//==========================================================
// scan_keys(...)
//  Compares the key state buffer to the MIDI key state.
//  If different, sends the appropriate key on/key off
//  messages.
//==========================================================
static bool scan_keys() {
    uint16 pitch;
    bool has_change = false;

    //Loop through the pitches.
    for(pitch = 0; pitch < 13; ++pitch) {

        //Get the state delta
        uint16 key = key_state[pitch];
        uint16 midi = midi_state[pitch];
        uint16 change = key ^ midi;

        //If state has changed, process it.
        if(change) {
            has_change = true;
            uint16 bitmask;
            uint16 octave = 0;
            for(bitmask = 1; bitmask; bitmask <<= 1, ++octave) {
                if(change & bitmask) {
                    //Note state has changed - determine whether it turned on or off
                    if(midi & bitmask) {
                        //New value = 0 (off)
                        midi_send_note_on( fx_config.channel[octave], lookup_midi_pitch(pitch, octave), 0 );
                    } else {
                        //New value = 1 (on)
                        midi_send_note_on( fx_config.channel[octave], lookup_midi_pitch(pitch, octave), 127 );
                    }
                }
            }
            midi_state[pitch] ^= change;
        }
    }
    return has_change;
}

//==========================================================
// translate_analog(...)
//  Converts a 10-bit analog value into a MIDI data value
//  (0..127), pre-scaled to the given 10-bit maximum.
//==========================================================
static char translate_analog(uint32 input, uint16 max) {
    //Adjust for hysteresis window
    max -= HYSTERESIS_MAGNITUDE;

    //output = (input * 0x7F) / max
    input *= 0x7F;
    input /= max;
    return (input <= 0x7F) ? input : 0x7F;
}

//==========================================================
// send_analog_message(...)
//  Sends the corresponding MIDI message for an analog value
//  given the current analog scan phase
//==========================================================
static void send_analog_message(uint16 value) {
    analog_config* config = &(fx_config.analog_map[analog_phase]);
    byte midi_channel = config->message_type;
    char translated_value = translate_analog(value, config->max_value);
    if(translated_value != last_translated_value[analog_phase]) {
        last_translated_value[analog_phase] = translated_value;
        switch(midi_channel & 0xF0) {
            case 0xB0:
                midi_send_controller(midi_channel & 0xF, config->message_data, translated_value);
                break;
            case 0xD0:
                midi_send_channel_aftertouch(midi_channel & 0xF, translated_value);
                break;
        }
    }
}

//==========================================================
// scan_analog()
//  Scans the four analog inputs.
//  If any have changed, sends the appropriate MIDI message.
//  Stops after one MIDI message is sent or after all four
//   inputs have been scanned with no change detected.
//==========================================================
static bool scan_analog() {
    int i;
    uint16 value;
    uint16 baseline;

    //Scan through four analog signals; stop if one has changed outside of hysteresis
    for(i = 0; i < 4; ++i) {
        //Advance to next sensor
        ++analog_phase;
        analog_phase &= 3;

        value = (&ADC1BUF0)[analog_phase];
        baseline = midi_analog_hysteresis[analog_phase];
        if(value < baseline) {
            midi_analog_hysteresis[analog_phase] = value;
            send_analog_message(value);
            return true;
        } else if(value > baseline + HYSTERESIS_MAGNITUDE) {
            midi_analog_hysteresis[analog_phase] = value - HYSTERESIS_MAGNITUDE;
            send_analog_message(value);
            return true;
        }
    }
    return false;
}

//==========================================================
// handle_continuous_change(...)
//  Sends the configured MIDI message for a slider or lever
//==========================================================
static void handle_continuous_change(uint16 control_index, byte new_value) {
    control_config* control = &(fx_config.control_map[control_index]);
    switch(control->message & 0xF0) {
        case MIDI_NOTE_ON:  //On if >=threshold; off if <threshold.  data1=pitch, data2=threshold (0 means new_value=velocity)
        case MIDI_CONTROLLER:  //On if >=threshold; off if <threshold.  data1=controller#, data2=threshold (0 means new_value=velocity)
            if(control->data2) {
                new_value = (new_value >= control->data2) ? 127 : 0;
            }
            midi_send_message_2(control->message, control->data1, new_value);
            break;

        case MIDI_POLY_AFTERTOUCH:  //data1=pitch, new_value=velocity
            midi_send_message_2(control->message, control->data1, new_value);
            break;

        case MIDI_PATCH_CHANGE:  //trigger if >= threshold.  data1=patch#, data2=threshold
            if(new_value >= control->data2)
                midi_send_message_1(control->message, control->data1);
            break;

        case MIDI_CHANNEL_AFTERTOUCH:  //new_value=velocity
            midi_send_message_1(control->message, new_value);
            break;

        //case MIDI_PITCH_BEND:  //Pitch bend: new_value=range (use same value for MSB, LSB)
    }
}

//==========================================================
// handle_slider_change(...)
//  Sends the configured MIDI message for a slider
//==========================================================
static void handle_slider_change(uint16 control_index, byte new_value) {
    //Value ranges from 1 to 15; translate to controller 0..127 by array lookup
    handle_continuous_change(control_index, slider_value_map[new_value & 15]);
}

//==========================================================
// handle_lever_change(...)
//  Sends the configured MIDI message for a lever
//==========================================================
static void handle_lever_change(uint16 control_index, byte new_value) {
    //Value ranges from 1 to 15; translate to controller 0..127 by array lookup
    handle_continuous_change(control_index, lever_value_map[new_value & 7]);
}

//==========================================================
// handle_momentary_button(...)
//  Sends the configured "on" or "off" MIDI message for a
//  momentary button
//==========================================================
static void handle_momentary_button(control_config* control, bool new_value) {
    switch(control->message & 0xF0) {
        case MIDI_NOTE_OFF: //Only send on button trigger; data1=pitch, data2=release velocity
            if(new_value)
                midi_send_message_2(control->message, control->data1, control->data2);
            break;

        case MIDI_NOTE_ON:  //data1=pitch, data2=velocity (fixed to 0 on button release)
        case MIDI_CONTROLLER:   //data1=controller; data2 = on value (off value = 0)
            midi_send_message_2(control->message, control->data1, (new_value ? control->data2 : 0));
            break;

        case MIDI_PATCH_CHANGE: //Only send on button trigger; data1=patch#
            if(new_value) midi_send_message_1(control->message, control->data1);

        //case MIDI_POLY_AFTERTOUCH:
        //case MIDI_CHANNEL_AFTERTOUCH:
        //case MIDI_PITCH_BEND:
    }
}

//==========================================================
// handle_button_change(...)
//  Sends the configured MIDI message for a button whose
//  state has changed.
//==========================================================
static void handle_button_change(uint16 control_index, bool new_value) {
    control_config* control = &(fx_config.control_map[control_index]);
    switch(control->control_type & 0xF0) {
        case subtype_momentary:
            handle_momentary_button(control, new_value);
            break;

        case subtype_persistent:
            if(new_value) handle_momentary_button(control, true);
            break;

        //case subtype_toggle:  //Requires special logic
        //case subtype_progressive: //Requires special logic
    }
}

//==========================================================
// scan_cpanel()
//  Compares the 65-word control panel buffer to the MIDI
//   control panel buffer.
//  If a bit is different, sends the appropriate message.
//  Optimized via the vector table, since the scan iterates
//   over the control map and not the buffer.
//==========================================================
static void scan_cpanel() {
    //Control panel has changed, so scan it.
    //
    //Starting with the first control in fx_config:
    // * Get word index from bit index
    // * Compute delta bits in word
    // * Get next control index (J) from vector table
    // * If delta bits are nonzero, then for each control in the range of current to J-1:
    //      - If control is a lever or slider:
    //          > Extract the lever/slider delta
    //          > If nonzero, translate and handle the lever/slider message
    //      - Else
    //          > Extract the Boolean delta
    //          > If changed, handle the button/LED message
    // * If next control index is nonzero, skip to the next control index; otherwise stop.

    //Assumptions:
    // * Controls are in ascending order of bit index
    // * Terminating control record is present
    // * Vector table is correctly populated

    byte control_index = 0;

    do {
        //Get values from tables
        byte next_control_index = fx_cpanel_skipvector[control_index];
        uint16 word_index = fx_config.control_map[control_index].bit_index >> 4;
        if(word_index < 65) {
            
            //Control is valid.  Now compute delta bits.
            uint16 new_value = cpanel_buffer[word_index];
            uint16 delta = new_value ^ midi_cpanel_state[word_index];
            if(delta != 0) {
                //State has changed.  First save new value for accounting, then send MIDI messages.
                midi_cpanel_state[word_index] = new_value;

                while(control_index < next_control_index) {
                    uint16 delta_value;
                    uint16 bit_index = fx_config.control_map[control_index].bit_index & 15;

                    switch(fx_config.control_map[control_index].control_type & 0xF) {
                        case type_slider: //Bit index 0, 4, 8 or 12
                            delta_value = (delta >> (12 - bit_index)) & 15;
                            if(delta_value != 0) {
                                //Slider changed
                                handle_slider_change(control_index, ((new_value >> (12-bit_index)) & 15));
                            }
                            break;

                        case type_lever: //Bit index 0, 4, 8 or 12
                            delta_value = (delta >> (12 - bit_index)) & 7;
                            if(delta_value != 0) {
                                //Lever changed
                                handle_lever_change(control_index, ((new_value >> (12-bit_index)) & 7));
                            }
                            break;

                        default:
                            if(((delta << bit_index) & 0x8000) != 0) {
                                //Button changed
                                handle_button_change(control_index, ((new_value << bit_index) & 0x8000));
                            }
                            break;
                    }
                    ++control_index;
                }
            }
        }

        if(next_control_index > control_index)
            control_index = next_control_index;
        else
            control_index = 0;
    } while(control_index != 0);
}

//==========================================================
// midi_scan()
//  Main loop - waits for input events and sends the
//   appropriate MIDI messages for the events.
//  Keyboard scanning has highest priority.
//  If no keys have changed, then the analog inputs are
//   scanned, stopping after the first change.
//  If the interrupts have flagged a change in the control
//   panel data, the control panel is scanned.
//==========================================================
void midi_scan() {
    while(true) {
        if(!scan_keys()) {
            scan_analog();
        }
        if(fx_cpanel_changed) {
            fx_cpanel_changed = false;
            scan_cpanel();
        }
    }
}

