/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * shiftreg.c:  Shift register operation for keyboard scan
 **********************************************************/
#include "common.h"
#include "shiftreg.h"

// IMPORTANT NOTE:
//  Programmer cannot be plugged in at same time as MIDI when installed in the organ.
//  The programmer increases the capacitance on the shift register clock line
//  so that no shift occurs.

//==========================================================
// Globals
//==========================================================
uint16 spi_dma_buffer[4];

//==========================================================
// shiftreg_init()
// Configures the SPI1 and DMA0 modules
//==========================================================
void shiftreg_init() {
    //Init SPI1 module
    // 1 MHz clock
    // Master mode, input only
    // Clock idle is low
    // Input logic 0 is low
    // Sample input on falling clock edge
    // Pin assignments:
    //   - Clock (SCK1):   Non-remappable
    //   - Data In (SDI1): Non-remappable
    //	 - Shift/-Load:    Pin 14 (RB5)

    _LATB5 = 0;
    _TRISB5 = 0;

    SPI1CON1 = 0x0D2B;  //800 = SDO disabled
                        //400 = 16-bit mode
                        //200 = Data sampled at end of data output (0 = middle)
                        //100 = Output on toward-idle clock edge (0 = toward-active)
                        // 20 = Master mode
                        // 3 bits secondary prescale (0=8:1 ... 7=1:1)
                        // 2 bits primary prescale (3/2/1/0 = 1/4/16/64:1)

    //010 11 = (6*1):1 = 10 MHz @ 60 MIPS (takes about 1/2 scan phase to read 32 bits)

    _SPIEN = 1;

    DMA0PAD = (uint16)(&SPI1BUF);
    DMA0REQ = 0xA; //SPI1 transfer done
    DMA0STAH = 0;
    _DMA0IP = 5;
    _DMA0IF = 0;
    _DMA0IE = 1;
    DMA0CON = 0x8801; //Enabled, read word, write null, one-shot, single buffer
}

//==========================================================
// shiftreg_load_beginread_dma()
// Triggers the shift register parallel input, then
//  begins a 2-word DMA read of the shift register.
// The DMA0 interrupt fires when the read is completed.
// Return value is the state of PORTB as read during the
//  parallel input operation.
//==========================================================
uint16 shiftreg_load_beginread_dma() {
    uint16 rb_value;

    _LATB5 = 0;
    Nop(); 	Nop(); 	Nop(); 
    rb_value = PORTB;
    _LATB5 = 1;

    //Trigger a 2-word read
    _DMA0IF = 0;
    DMA0STAL = (uint16)(&spi_dma_buffer);
    DMA0CNT = 1;
    DMA0CON = 0x8801;
    SPI1BUF = 0;

    return rb_value;
}

