/**********************************************************
 * MIDI Out Module for Yamaha Electone FX/FS Series organs
 * Tim Singer, 2012
 * Project page: http://tsstech.org/electronics/fxmidi.html
 * 
 * uart.c:  RS-232 communication
 *
 **********************************************************/
#include "common.h"
#include "uart.h"

volatile int command = 0;

//==========================================================
// UART transmit queue

static volatile char _queue[256];
static volatile byte _head;
static volatile byte _tail;

//==========================================================
// Initialization

#define U1TX 1

void uart_init() {

	//Initialize locals
	_head = _tail = 0;

	//Initialize the hardware
	// * UART1, transmit only
	// * 8 data bits, 1 stop bit, no parity
	// * ~115200 baud (60MHz / (4 * 130))
	// * Negative RS232 voltage (pin at +5V) is idle

	UART_TX_REMAP_PIN_REG = U1TX;	//Assign pin to UART 1 TX
	_U1RXR = UART_RX_REMAP_PIN;		//Assign pin to UART 1 RX
	U1BRG = 129;
	U1MODE = 0x8808;
	U1STA = 0x8400;		//Set after mode, per reference manual	

	_U1TXIP = 2;
	_U1TXIF = 0;
	_U1TXIE = 1;

	_U1RXIP = 1;
	_U1RXIF = 0;
	_U1RXIE = 1;
}

//==========================================================
// UART queue functions

#define queue_empty() (_head == _tail)
#define queue_full() ( (byte)(_tail + 1) == _head )	//Use integer overflow

//==========================================================
// UART TX functions

void uart_send_byte(char data) {
	//Enqueue the data
	while(queue_full()) { }
	_queue[_tail] = data;
	++_tail;

	//Trigger the interrupt to start the transmission
	if(!U1STAbits.UTXBF)
		_U1TXIF = 1;
}

void uart_send_bytes(const char* data, uint16 count) {
	while(count) {
		uart_send_byte(*data);
		++data;
	}
}

void uart_send_string(const char* data) {
	while(*data) {
		uart_send_byte(*data);
		++data;
	}
}

int print_digit_opt(uint16 i, int print) {
	char c;
	c = (char)(i%10);
	if(print || c) {
		uart_send_byte('0' + c);
		return 1;
	}
	return 0;
}

void uart_print_int(uint16 i) {
	//if(i == 0) uart_send_byte('0');

	int print = print_digit_opt(i / 10000, 0);
	print = print_digit_opt(i / 1000, print);
	print = print_digit_opt(i / 100, print);
	print_digit_opt(i / 10, print);
	print_digit_opt(i, 1);
}

void print_hex_digit(uint16 i) {
	char c;
	c = ((char)i);
	if(c < 10)
		uart_send_byte(c + '0');
	else
		uart_send_byte(c + ('a' - 10));
}

void uart_print_hex(uint16 i) {
	print_hex_digit((i&0xF000)>>12);
	print_hex_digit((i&0xF00)>>8);
	print_hex_digit((i&0xF0)>>4);
	print_hex_digit(i&0xF);
}

void uart_print_hex8(byte i) {
	print_hex_digit((i&0xF0)>>4);
	print_hex_digit(i&0xF);
}

//==========================================================
// UART transmit interrupt

void __attribute__((__interrupt__, no_auto_psv)) _U1TXInterrupt(void) {
	_U1TXIF = 0; // Clear TX Interrupt flag

	while(!U1STAbits.UTXBF) {
		if(queue_empty()) break;
		U1TXREG = _queue[_head];
		++_head;
	}
}

//==========================================================
// UART receive interrupt

extern volatile bool logic_probe_cancel;

void __attribute__((__interrupt__, no_auto_psv)) _U1RXInterrupt(void) {
	_U1RXIF = 0; // Clear RX Interrupt flag
	char c = (char)U1RXREG;

	//Command 'r' / 0x80: Reset into bootloader for firmware update
	if(c == (char)0x80 || c == 'r') {
		asm volatile("reset");
    } else if(c == 'p') {
        command = 'p';
    } else if(c == 's') {
        //logic_probe_cancel = true;
    }
}
